/* rpi.c - Provides routines specific to memu-pi */

#include "types.h"
#include "diag.h"
#include "config.h"
#include "common.h"
#include "gpio.h"

#include <unistd.h>
/* #include <linux/reboot.h> */
#include <sys/reboot.h>
#include <sys/mount.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
#include <termios.h>
#include <stdarg.h>

BOOLEAN	bOS	=  FALSE;	/* True if running under a full OS */

static FILE *pfilDbg =	NULL;

int memu_entry (int argc, char *argv[]);
void set_gpu_mode (int);
int get_gpu_mode (void);

/* A replacement for the normal exit () system call, which halts the system */

void rpi_exit (int iStatus)
   {
   int iExit;
   /* Restore the GPIO */

   hw_term ();

   /* Exit to OS if available */

   if ( bOS )  exit (iStatus);

   /* Sync the file system to ensure all pending writes have completed */

   printf ("Memu-Pi shutting down.\n");
   sync ();
   sleep (2);
   sync ();
   sleep (2);
   printf ("Filesystem synced.\n");

   /* Give time to read error message */

   if ( iStatus != 0 )  sleep (10);

   /* Halt the system */

   // printf ("Shutting down\n");
   iExit = reboot (RB_POWER_OFF);
   perror ("reboot: ");

   /* Failed to power down - wait forever */

   while (TRUE) sleep (1000000);
   }

static void uk_kbd_map (void)
	{
	static int ttyfd  =  0;
	struct kbentry kbe;
	struct
		{
		int	code;
		int shift;
		int key;
		} uk_map[] = {
		{ 0x03, K_SHIFTTAB, '"' },
			{ 0x28,  K_NORMTAB, '\'' },
			{ 0x28, K_SHIFTTAB, '@' },
			{ 0x2b, K_NORMTAB, '#' },
			{ 0x2b, K_SHIFTTAB, '~' },
			{ 0x56, K_NORMTAB, '\\' },
			{ 0x56, K_SHIFTTAB, '|' } };
	int i;

	for ( i = 0; i < sizeof (uk_map) / sizeof (uk_map[0]); ++i )
		{
		kbe.kb_table   =  uk_map[i].shift;
		kbe.kb_index   =  uk_map[i].code;
		kbe.kb_value   =  uk_map[i].key;
		if ( ioctl (ttyfd, KDSKBENT, &kbe) < 0 )
			diag_message (DIAG_HW_KBD, "Failed to map %s key 0x%02x to \'%c\'",
			( uk_map[i].shift == K_SHIFTTAB ? "Shifted" : "Normal" ), uk_map[i].code, uk_map[i].key);
		}
	}

static void mount_proc (void)
	{
	if ( mount ("proc", "/proc", "proc", MS_NODEV | MS_NOEXEC | MS_NOSUID, NULL) != 0 )
		{
		fatal ("Failed to mount proc filesystem.");
		}
	}

int main (int argc, char *argv[])
	{
	int	i;
	/* Read configuration file */
	// dbg_message ("argv[0] = %s\n", argv[0]);
	char *psCfg = cfg_find_file (argv[0]);
    // dbg_message ("psCfg = %s\n", psCfg);
	if ( psCfg == NULL )
		{
		// dbg_message ("Trying default configuration file.\n");
		psCfg = estrdup ("/memu/memu-pi.cfg");
		if ( ! cfg_test_file (psCfg) ) fatal ("No configuration file found");
		}
	if ( ! read_config (psCfg, &argc, &argv) ) fatal ("Error reading configuration file");
	// dbg_message ("Read configuration file.\n");
	config_set_file (psCfg);

	/* Test for OS switch */

	for ( i = 0; i < argc; ++i )
		{
		if ( ! strcmp (argv[i], "-os") )	bOS	  =	 TRUE;
		if ( ! strcmp (argv[i], "-noos") )	bOS	  =	 FALSE;
		}
	// dbg_message ("bOS = %d\n", bOS);

	if ( ! bOS )
		{
		/* Configure UK keyboard */
		// dbg_message ("Applying UK key map.\n");
		uk_kbd_map ();
		/* Mount /proc (in order to be able to determine RPi revision) */
		// dbg_message ("Mounting proc filesystem.\n");
		mount_proc ();
		/* Wait for end of kernel messages and then start MEMU */
		sleep (2);
		}

	/* printf ("Calling memu_entry.\n"); */
	// dbg_message ("Entering main MEMU code.\n");
	memu_entry (argc, argv);
	}

void rpi_init (void)
	{
//	int	iRev;

	/* Configure GPIO to read joysticks */

	// dbg_message ("Reached rpi_init.\n");
	diag_message (DIAG_ALWAYS, "MEMU-Pi: bOS = %d", bOS);
	hw_init ();
	}

BOOLEAN rpi_options (int *pargc, char ***pargv, int *pi)
	{
	if ( !strcmp((*pargv)[*pi], "-os") )
		{
		/* A no-op at this stage - has already been detected */
		}
	else if ( !strcmp((*pargv)[*pi], "-noos") )
		{
		/* Ditto */
		}
	else if ( !strcmp((*pargv)[*pi], "-hw-config") )
		{
		if ( ++(*pi) == (*pargc) )
			usage();
		hw_read ((*pargv)[*pi]);
		}
	else if ( !strcmp((*pargv)[*pi], "-gpu-mode") )
		{
		if ( ++(*pi) == (*pargc) )
			usage();
		set_gpu_mode (atoi ((*pargv)[*pi]));
		}
	else if ( ! cfg_options (pargc, pargv, pi) )
		{
		return FALSE;
		}
	return	TRUE;
	}

void rpi_usage (void)
	{
	fprintf(stderr, "       -os                  running under an operating system\n");
	fprintf(stderr, "       -noos                running as init task\n");
	fprintf(stderr, "       -hw-config file      name of file specifying hardware\n");
	fprintf(stderr, "       -gpu-mode            mode for scaling window to display\n");
	cfg_usage ();
	}

void rpi_save_cfg (FILE *pfil)
	{
	if ( bOS ) fprintf (pfil, "-os\n");
	else fprintf (pfil, "-noos\n");
	int mode = get_gpu_mode ();
	if ( mode ) fprintf (pfil, "-gpu-mode %d\n", mode);
	hw_save_cfg (pfil);
	}
