/*	tape.c	-	Cassette tape hardware emulation */

#include "common.h"
#include "tape.h"
#include "ctc.h"
#include "diag.h"
#include "memu.h"
#include <stdio.h>
#include <string.h>

#define	MTX_TAPE_ZERO	0x348			/* Duration of a zero bit half cycle (in Z80 clocks) */
#define	MTX_TAPE_ONE	0x690			/* Duration of a one bit half cycle (in Z80 clocks) */
#define MTX_TAPE_HALF	( ( MTX_TAPE_ZERO + MTX_TAPE_ONE ) / 2 )

typedef enum { fmtUnk, fmtMTX, fmtCAS, fmtWAV, fmtW08, fmtW16, fmtW32 } TapeFmt;
typedef enum { stgPre, stgWait, stgData } MTXStages;

static BOOLEAN bTapeRun = FALSE;		/* Tape running */

static const char *psTapeIn = NULL;		/* Input file name */
static TapeFmt fmtIn = fmtUnk;			/* Input file format */
static FILE *pfTapeIn = NULL;			/* Input file stream */
static int cTapeNext = 0;				/* Z80 clocks until next interrupt */
static MTXStages stgIn = stgPre;		/* Stages of loading an MTX tape file */
static int nCycleIn;					/* Number of half cycles in tape preamble */
static BOOLEAN bMTXFirst;				/* First half cycle of an MTX tape bit */
static int nMTXBitsIn;					/* Number of bits remaining in MTX tape byte */
static int nMTXBytesIn;					/* Number of bytes loaded in a stage */
static byte byMTXIn;					/* MTX tape byte */
static unsigned int cRes = 0;		   	/* Z80 clocks remaining from last tape advance */
static word nChan;						/* Number of channels in WAV file */
static unsigned int nRateIn;			/* WAV file sample rate */
static float fTapeLast;					/* Last value read from tape */
static int nPulse;						/* Number of CTC trigger pulses since tape last started */

static const char *psTapeOut = NULL;	/* Output filename */
static TapeFmt fmtOut = fmtUnk;			/* Output file format */
static FILE *pfTapeOut = NULL;			/* Output file stream */
static byte byTapeOut = 0;				/* Last value output to tape */
static unsigned int cTapeLast = 0;		/* Z80 clock at last tape output */
static MTXStages stgOut = stgPre;		/* Stages of saving an MTX tape file */
static int nMTXBitsOut;					/* Number of bits in MTX tape byte */
static int nMTXBytesOut;				/* Number of bytes saved to an MTX tape file */
static int nSysBytes;					/* Number of bytes of system variables */
static byte byMTXOut;					/* MTX tape output byte */
static byte nCycleOut;					/* Number of preamble cycles output */
static int nRateOut = 4800;				/* WAV output file sample rate */

extern CFG cfg;

static TapeFmt tape_identify (const char *psFile)
	{
	const char *psExt;
	int nCh = strlen (psFile);
	if ( nCh < 3 ) return fmtUnk;
	psExt = &psFile[nCh-3];
	if ( ! strcasecmp (psExt, "mtx") ) return fmtMTX;
	if ( ! strcasecmp (psExt, "cas") ) return fmtCAS;
	if ( ! strcasecmp (psExt, "wav") ) return fmtWAV;
	return fmtUnk;
	}

void tape_set_input (const char *psFile)
	{
	psTapeIn = psFile;
	fmtIn = tape_identify (psFile);
	if ( fmtIn == fmtUnk ) fatal ("Unknown tape in format");
	}

static float tape_read_wav (void)
	{
	if ( pfTapeIn == NULL ) return 0.0;
	float fVal = 0.0;
	int iChan;
	int nRead;
	byte by;
	short wd;
	float f;
	for ( iChan = 0; iChan < nChan; ++iChan)
		{
		switch (fmtIn)
			{
			case fmtW08:
				nRead = fread (&by, sizeof (by), 1, pfTapeIn);
				fVal += by / 128.0 - 1.0;
				break;
			case fmtW16:
				nRead = fread (&wd, sizeof (wd), 1, pfTapeIn);
				fVal += wd / 32768.0;
				break;
			case fmtW32:
				nRead = fread (&f, sizeof (f), 1, pfTapeIn);
				fVal += f;
				break;
			}
		if ( nRead == 0 )
			{
			fclose (pfTapeIn);
			pfTapeIn = NULL;
			diag_message (DIAG_TAPE, "End of tape: CTC Pulses = %d.", nPulse);
			return 0.0;
			}
		}
	return fVal / nChan;
	}

static void tape_open_cas (const char *psFile)
	{
	pfTapeIn = fopen (psFile, "rb");
	cTapeNext = MTX_TAPE_ONE;
	}

static void tape_open_mtx (const char *psFile)
	{
	pfTapeIn = fopen (psFile, "rb");
	cTapeNext = MTX_TAPE_ONE;
	stgIn = stgPre;
	}

static void tape_open_wav (const char *psFile)
	{
	unsigned int nLen;
	char sID[5];
	pfTapeIn = fopen (psFile, "rb");
	sID[4] = '\0';
	fread (sID, sizeof (char), 4, pfTapeIn);
	if ( strcmp (sID, "RIFF") )	fatal ("Tape WAV file does not have a RIFF header");
	fread (&nLen, sizeof (nLen), 1, pfTapeIn);
	fread (sID, sizeof (char), 4, pfTapeIn);
	if ( strcmp (sID, "WAVE") ) fatal ("Tape WAV file is not WAVE format");
	while ( fread (sID, sizeof (char), 4, pfTapeIn) == 4 )
		{
		fread (&nLen, sizeof (nLen), 1, pfTapeIn);
		if ( ! strcmp (sID, "fmt ") )
			{
			word wFmt;
			unsigned int nBPS;
			word wAlign;
			unsigned int nBits;
			word wBits;
			fread (&wFmt, sizeof (wFmt), 1, pfTapeIn);
			if ( wFmt != 1 ) fatal ("Tape WAV file is compressed");
			fread (&nChan, sizeof (nChan), 1, pfTapeIn);
			fread (&nRateIn, sizeof (nRateIn), 1, pfTapeIn);
			fread (&nBPS, sizeof (nBPS), 1, pfTapeIn);
			fread (&wAlign, sizeof (wAlign), 1, pfTapeIn);
			if ( nLen == 16 )
				{
				fread (&wBits, sizeof (wBits), 1, pfTapeIn);
				nBits = wBits;
				}
			else if ( nLen == 18 )
				{
				fread (&nBits, sizeof (nBits), 1, pfTapeIn);
				}
			else
				{
				fatal ("Tape WAV file has unexpected fmt chunk length");
				}
			if ( nBits == 8 )       fmtIn = fmtW08;
			else if ( nBits == 16 )	fmtIn = fmtW16;
			else if ( nBits == 32 ) fmtIn = fmtW32;
			else fatal ("Tape WAV file has unsupported sample length");
			}
		else if ( ! strcmp (sID, "data") )
			{
			if ( fmtIn == fmtWAV ) fatal ("No fmt chunk in WAV file before data");
			fTapeLast = tape_read_wav ();
			return;
			}
		else
			{
			fseek (pfTapeIn, nLen, SEEK_CUR);
			}
		}
	fatal ("No data chunk found in Tape WAV file");
	}

void tape_play (void)
	{
	if ( ( pfTapeIn == NULL ) && ( psTapeIn != NULL ) )
		{
		switch (fmtIn)
			{
			case fmtMTX:
				tape_open_mtx (psTapeIn);
				break;
			case fmtCAS:
				tape_open_cas (psTapeIn);
				break;
			default:
				tape_open_wav (psTapeIn);
				break;
			}
		}
	}

void tape_advance_cas (int clks)
	{
	// diag_message (DIAG_TAPE, "Tape Advance: clks = %d, cTapeNext = %d", clks, cTapeNext);
	cTapeNext -= clks;
	if ( cTapeNext <= 0 )
		{
		word cNext;
		// diag_message (DIAG_TAPE, "Tape interrupt.");
		++nPulse;
		ctc_trigger (3);
		if ( fread (&cNext, sizeof (cNext), 1, pfTapeIn) == 1 )
			{
			cTapeNext += cNext;
			}
		else
			{
			fclose (pfTapeIn);
			pfTapeIn = NULL;
			diag_message (DIAG_TAPE, "End of tape.");
			}
		}
	}

void tape_advance_mtx (int clks)
	{
	// diag_message (DIAG_TAPE, "Tape Advance: clks = %d, cTapeNext = %d", clks, cTapeNext);
	cTapeNext -= clks;
	if ( cTapeNext <= 0 )
		{
		// diag_message (DIAG_TAPE, "Tape interrupt.");
		++nPulse;
		ctc_trigger (3);
		switch (stgIn)
			{
			case stgPre:
				nCycleIn = 1500;
				++stgIn;
				diag_message (DIAG_TAPE, "Starting preamble.");
				/* Fall through to next case */
			case stgWait:
				cTapeNext = MTX_TAPE_ZERO;
				--nCycleIn;
				if ( nCycleIn == 0 )
					{
					cTapeNext = MTX_TAPE_ONE;
					diag_message (DIAG_TAPE, "Starting data.");
					++stgIn;
					nMTXBytesIn = 0;
					nMTXBitsIn = 0;
					bMTXFirst = TRUE;
					}
				break;
			case stgData:
				if ( bMTXFirst )
					{
					bMTXFirst = FALSE;
					}
				else
					{
					bMTXFirst = TRUE;
					byMTXIn >>= 1;
					--nMTXBitsIn;
					if ( nMTXBitsIn <= 0 )
						{
						if ( fread (&byMTXIn, sizeof (byMTXIn), 1, pfTapeIn) != 1 )
							{
							fclose (pfTapeIn);
							pfTapeIn = NULL;
							diag_message (DIAG_TAPE, "End of tape.");
							return;
							}
						// diag_message (DIAG_TAPE, "Tape byte = 0x%02x", byMTXIn);
						++nMTXBytesIn;
						nMTXBitsIn = 8;
						}
					}
				if ( byMTXIn & 0x01 ) 	cTapeNext = MTX_TAPE_ONE;
				else 					cTapeNext = MTX_TAPE_ZERO;
				// diag_message (DIAG_TAPE, "Set cTapeNext = %d", cTapeNext);
			}
		}
	}

void tape_advance_wav (int clks)
	{
	BOOLEAN bInt = FALSE;
	int nSamp = nRateIn * ( clks + cRes ) / 4000000;
	int iSamp;
	cRes += clks - 4000000 * nSamp / nRateIn;
	for ( iSamp = 0; iSamp < nSamp; ++iSamp )
		{
		float fTapeIn = tape_read_wav ();
		if ( ( ( fTapeIn > 0.0 ) && ( fTapeLast <= 0.0 ) )
			|| ( ( fTapeIn < 0.0 ) && ( fTapeLast >= 0.0 ) ) ) bInt = TRUE;
		fTapeLast = fTapeIn;
		}
	if ( bInt )
		{
		// diag_message (DIAG_TAPE, "Tape interrupt.");
		++nPulse;
		ctc_trigger (3);
		}
	}

void tape_advance (int clks)
	{
	if ( ( bTapeRun ) && ( pfTapeIn != NULL ) )
		{
		switch (fmtIn)
			{
			case fmtMTX:
				tape_advance_mtx (clks);
				break;
			case fmtCAS:
				tape_advance_cas (clks);
				break;
			default:
				tape_advance_wav (clks);
				break;
			}
		}
	}

void tape_out1F (byte value)
	{
	static int	isave;
	switch ( value )
		{
		case 0xaa:
			diag_message (DIAG_TAPE, "Tape start.");
			nPulse = 0;
			tape_play ();
			bTapeRun = TRUE;
			isave = cfg.iperiod;
			if ( fmtIn >= fmtWAV ) 	cfg.iperiod = 105;
			else					cfg.iperiod = 840;
			if ( fmtOut == fmtMTX ) stgOut = stgPre;
			if ( fmtIn == fmtMTX )
				{
				stgIn = stgPre;
				cTapeNext = MTX_TAPE_ZERO;
				}
			break;
		case 0x55:
			diag_message (DIAG_TAPE, "Tape stop: CTC Pulses = %d.", nPulse);
			bTapeRun = FALSE;
			cfg.iperiod = isave;
			if ( fmtIn == fmtMTX )
				{
				word addr;
				diag_message (DIAG_TAPE, "%d Bytes loaded, %d Bits remaining.", nMTXBytesIn, nMTXBitsIn);
/*
				Z80 *r = get_Z80_regs ();
				printf ("HL = 0x%04x\n", r->HL.W);
				addr = r->HL.W - 20;
				while ( addr <= r->HL.W )
					{
					printf (" 0x%02x", mem_read_byte (addr));
					++addr;
					}
				printf ("\n");
				fatal ("End of test");
*/
				}
			if ( fmtOut == fmtMTX )
				{
				diag_message (DIAG_TAPE, "%d data bytes saved.", nMTXBytesOut);
				}
			break;
		}
	}

void tape_set_output (const char *psFile)
	{
	psTapeOut = psFile;
	fmtOut = tape_identify (psFile);
	if ( fmtOut == fmtUnk ) fatal ("Unknown tape out format");
	// diag_message (DIAG_TAPE, "Set cassette output file %s", psFile);
	}

void tape_save_cas (byte value)
	{
	unsigned int	cNow = get_Z80_clocks ();
	word cTime;
	if ( pfTapeOut == NULL )
		{
		if ( psTapeOut == NULL ) OutZ80_bad("cassette tape", 0x03, value, FALSE);
		pfTapeOut = fopen (psTapeOut, "wb");
		}
	cTime = cNow - cTapeLast;
	fwrite (&cTime, sizeof (cTime), 1, pfTapeOut);
	cTapeLast = cNow;
	}

void tape_save_mtx (byte value)
	{
	unsigned int	cNow = get_Z80_clocks ();
	word cTime = cNow - cTapeLast;
	cTapeLast = cNow;
	switch (stgOut)
		{
		case stgPre:
			nCycleOut = 0;
			++stgOut;
			break;
		case stgWait:
			++nCycleOut;
			if ( ( value ) && ( cTime > MTX_TAPE_HALF ) )
				{
				if ( pfTapeOut == NULL )
					{
					if ( psTapeOut == NULL ) OutZ80_bad("cassette tape", 0x03, value, FALSE);
					pfTapeOut = fopen (psTapeOut, "wb");
					}
				diag_message (DIAG_TAPE, "%d Preamble cycles", nCycleOut);
				nMTXBytesOut = 0;
				nMTXBitsOut = 0;
				byMTXOut = 0x00;
				bMTXFirst = TRUE;
				++stgOut;
				}
			break;
		case stgData:
			if ( value )
				{
				byMTXOut >>= 1;
				if ( cTime > MTX_TAPE_HALF ) byMTXOut |= 0x80;
				if ( ++ nMTXBitsOut >= 8 )
					{
					fwrite (&byMTXOut, sizeof (byMTXOut), 1, pfTapeOut);
					++nMTXBytesOut;
					nMTXBitsOut = 0;
					byMTXOut = 0x00;
					}
				}
		}
	cTapeLast = cNow;
	}

void tape_save_wav (byte value)
	{
	unsigned int	cNow = get_Z80_clocks ();
	word cTime = cNow - cTapeLast;
	cTapeLast = cNow;
	if ( cTime > 2 * MTX_TAPE_ONE ) cTime = 2 * MTX_TAPE_ONE;
	if ( pfTapeOut == NULL )
		{
		int iVal;
		word wVal;
		pfTapeOut = fopen (psTapeOut, "wb");
		fwrite ("RIFF", sizeof (char), 4, pfTapeOut);
		nMTXBytesOut = 0;
		fwrite (&nMTXBytesOut, sizeof (nMTXBytesOut), 1, pfTapeOut);
		fwrite ("WAVE", sizeof (char), 4, pfTapeOut);
		fwrite ("fmt ", sizeof (char), 4, pfTapeOut);
		iVal = 18;	/* Chunk size */
		fwrite (&iVal, sizeof (iVal), 1, pfTapeOut);
		wVal = 1;	/* Format */
		fwrite (&wVal, sizeof (wVal), 1, pfTapeOut);
		wVal = 1;	/* Number of channels */
		fwrite (&wVal, sizeof (wVal), 1, pfTapeOut);
		fwrite (&nRateOut, sizeof (nRateOut), 1, pfTapeOut);
		fwrite (&nRateOut, sizeof (nRateOut), 1, pfTapeOut);
		wVal = 1;	/* Block align */
		fwrite (&wVal, sizeof (wVal), 1, pfTapeOut);
		iVal = 8;	/* Bits per sample */
		fwrite (&iVal, sizeof (iVal), 1, pfTapeOut);
		fwrite ("data", sizeof (char), 4, pfTapeOut);
		fwrite (&nMTXBytesOut, sizeof (nMTXBytesOut), 1, pfTapeOut);
		}
	else
		{
		int nSamp = ( cTime * nRateOut + 2000000 ) / 4000000;
		int iSamp;
		byte byOut = byTapeOut ? 255 : 0;
		for ( iSamp = 0; iSamp < nSamp; ++iSamp ) fwrite (&byOut, sizeof (byOut), 1, pfTapeOut);
		nMTXBytesOut += nSamp;
		}
	}

void tape_out3 (byte value)
	{
	value &= 0x01;
	// diag_message (DIAG_TAPE, "Cassette output 0x%02x", value);
	if ( ( bTapeRun ) && ( value != byTapeOut ) )
		{
		switch (fmtOut)
			{
			case fmtMTX:
				tape_save_mtx (value);
				break;
			case fmtCAS:
				tape_save_cas (value);
				break;
			default:
				tape_save_wav (value);
				break;
			}
		}
	byTapeOut = value;
	}

void tape_out_close (void)
	{
	if ( fmtOut == fmtWAV )
		{
		/* Output final bit */
		int nSamp = ( MTX_TAPE_ONE * nRateOut + 2000000 ) / 4000000;
		int iSamp;
		byte byOut = byTapeOut ? 255 : 0;
		for ( iSamp = 0; iSamp < nSamp; ++iSamp ) fwrite (&byOut, sizeof (byOut), 1, pfTapeOut);
		/* Size of data chunk */
		fseek (pfTapeOut, 42, SEEK_SET);
		fwrite (&nMTXBytesOut, sizeof (nMTXBytesOut), 1, pfTapeOut);
		/* Size of WAVE file */
		nMTXBytesOut += 34;
		fseek (pfTapeOut, 4, SEEK_SET);
		fwrite (&nMTXBytesOut, sizeof (nMTXBytesOut), 1, pfTapeOut);
		fseek (pfTapeOut, 0, SEEK_END);
		}
	fclose (pfTapeOut);
	pfTapeOut == NULL;
	}

void tape_term (void)
	{
	if ( pfTapeOut != NULL ) tape_out_close ();
	if ( pfTapeIn != NULL )
		{
		fclose (pfTapeIn);
		pfTapeIn = NULL;
		}
	}
